/*

*/

package language4a;

import tree.*;
import tokenizer.*;

/**
   Analyze (check) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.
   <pre>{@code
       Prog ::= Expr
              | '(' 'prog' Expr+ ')'

       Expr ::= If
              | While
              | Begin
              | Set
              | Var
              | Print
              | AExp
              | BExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

      Begin ::= '(' 'begin' Expr+ ')'

        Set ::= '(' 'set' VARIABLE Expr ')'

        Var ::= '(' 'var' VARIABLE Expr ')'

      Print ::= '(' 'print' Expr ')'

      BExpr ::= '(' '||'  Expr Expr+ ')'
              | '(' '&&'  Expr Expr+ ')'
              | '(' '!'   Expr ')'
              | '('  EqOp Expr Expr ')'
              | '(' RelOp Expr Expr ')'

       EqOp ::= '==' | '!='
      RelOp ::= '<' | '>' | '<=' | '>='

      AExpr ::= '(' '+' Expr Expr* ')'
              | '(' '-' Expr Expr? ')'
              | '(' '*' Expr Expr+ ')'
              | '(' '/' Expr Expr  ')'
              | '(' '%' Expr Expr  ')'
              | '(' '**' Expr Expr ')'

    INTEGER ::= [-|+][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
   }</pre>
*/
public class AnalyzeAST
{
   public static boolean thereIsAnError = false;

   /**
      The methods checkProg(), checkExpr(), checkAexpr(), checkBexpr(),
      and checkRexpr() are essentially a post-order traversal of the abstract
      syntax tree.

      @param tree  an expression {@link Tree} to be checked
      @return the {@link Value} of the expression {@link Tree}
   */
   public static Value check(final Tree<Token> tree)
   {
      // Instantiate a global environment object.
      final Environment env = new Environment();

      final Value value = checkProg( tree, env );

      return value;
   }


   /**
      Check a program using the given {@link Environment}.

      @param tree  an expression {@link Tree} to be checked
      @param env   an {@link Environment} for local variables
      @return the {@link Value} of the expression {@link Tree}
   */
   public static Value checkProg(final Tree<Token> tree,
                                 final Environment env)
   {
      thereIsAnError = false;

      final Value result;  // a blank final

      // Check which kind of Prog we have.
      if ( ! tree.root().lexeme.equals("prog") )
      {
         // Check the single expression.
         result = checkExpr( tree, env );
      }
      else
      {
         // Check each Expr in the Prog.
         for (int i = 0; i < tree.subTrees().size() - 1; ++i)
         {
            checkExpr( tree.subTrees().get(i), env );
         }
         // Check the last Expr in the Prog and use it as the type.
         result = checkExpr( tree.subTrees().get(tree.subTrees().size()-1), env );
      }

      if (thereIsAnError)
      {
         throw new AnalysisError();
      }
      return result;
   }


   // Check an expression.
   private static Value checkExpr(final Tree<Token> tree,
                                  final Environment env)
   {
      final Value result;  // a blank final

      final String node = tree.root().lexeme;

      if ( node.equals("if") )
      {
         result = checkIf( tree, env );
      }
      else if ( node.equals("while") )
      {
         result = checkWhile( tree, env );
      }
      else if ( node.equals("begin") )
      {
         result = checkBegin(tree, env);
      }
      else if ( node.equals("set") )
      {
         result = checkSet(tree, env);
      }
      else if ( node.equals("var") )
      {
         result = checkVar(tree, env);
      }
      else if ( node.equals("print") )
      {
         result = checkPrint(tree, env);
      }
      else if ( node.equals("&&")
             || node.equals("||")
             || node.equals("!") )
      {
         result = checkBexpr(tree, env);  // boolean expression
      }
      else if ( node.equals("==")
             || node.equals("!=") )
      {
         result = checkEqexpr(tree, env);  // equality operator
      }
      else if ( node.equals("<")
             || node.equals(">")
             || node.equals("<=")
             || node.equals(">=") )
      {
         result = checkRelexpr(tree, env);  // relational operator
      }
      else if ( node.equals("+")
             || node.equals("-")
             || node.equals("*")
             || node.equals("/")
             || node.equals("%")
             || node.equals("**") )
      {
         result = checkAexpr(tree, env);  // arithmetic expression
      }
      else if ( tree.subTrees().size() == 0 ) // a leaf node
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( node.equals("true") );
         }
         else if ( node.matches("^[-|+]*[0-9][0-9]*") )
         {
            result = new Value( Integer.parseInt( node ) );
         }
         else if ( env.defined(node) )  // a defined variable
         {
            result = env.lookUp( node );
         }
         else  // runtime check
         {
            thereIsAnError = true;
            System.err.println("Syntax error: undefined variable: " + node);
            result = new Value( false );
         }
      }
      else
      {
         thereIsAnError = true;
         System.err.println("Syntax error: invalid expression: " + toString(tree));
         result = new Value( false );
      }
      return result;
   }


   // Check an if-expression.
   private static Value checkIf(final Tree<Token> tree,
                                final Environment env)
   {



   }


   // Check a while-loop expression.
   private static Value checkWhile(final Tree<Token> tree,
                                   final Environment env)
   {



      // Always return false for a while-loop expression.
      return new Value(false);
   }


   // Check a begin expression.
   private static Value checkBegin(final Tree<Token> tree,
                                   final Environment env)
   {
      // Create a new Environment object chained to (or "nested in")
      // the previous ("outer") environment object.
      final Environment newEnv = new Environment(env);



   }


   // Check a set expression.
   private static Value checkSet(final Tree<Token> tree,
                                 final Environment env)
   {



   }


   // Check a var expression.
   private static Value checkVar(final Tree<Token> tree,
                                 final Environment env)
   {



   }


   // Check a print expression.
   private static Value checkPrint(final Tree<Token> tree,
                                   final Environment env)
   {



   }


   // Check a boolean expression.
   private static Value checkBexpr(final Tree<Token> tree,
                                   final Environment env)
   {




      return new Value(false); // boolean expressions always return a boolean
   }


   // Check an equality expression (which is a boolean expression).
   private static Value checkEqexpr(final Tree<Token> tree,
                                    final Environment env)
   {




      return new Value(false); // equality expressions always return a boolean
   }


   // Check a relational expression (which is a boolean expression).
   private static Value checkRelexpr(final Tree<Token> tree,
                                     final Environment env)
   {





      return new Value(false); // relational expressions always return a boolean
   }


   // Check an arithmetic expression.
   private static Value checkAexpr(final Tree<Token> tree,
                                   final Environment env)
   {




      return new Value(0); // arithmetic expressions always return an integer
   }


   /**
      Convert a {@code Tree} of {@link Token}s into a {@link String}.
      <p>
      This method does a pre-order traversal of the {@link Tree}.
      <p>
      This method is used by the analysis error messages.

      @param tree  {@link Tree} of {@link Token}s to convert into a {@link String}
      @return a {@link String} representation of this {@code Tree} of {@link Token}s
   */
   public static String toString(final Tree<Token> tree)
   {
      String result = "";
      if ( tree.isLeaf() )
      {
         result += tree.root().lexeme;
      }
      else
      {
         result += "(" + tree.root().lexeme;
         for ( final Tree<Token> subTree : tree.subTrees() )
         {
            result += " " + toString(subTree);  // recursion
         }
         result += ")";
      }
      return result;
   }
}
