/*

*/

package tree;

import java.io.PrintWriter;
import java.io.File;

/**
   This program converts a {@link Tree} data structure
   into a DOT language description of the {@link Tree}.
   The DOT description is written to a file with the
   .dot extension. The dot file can then be processed
   by the dot.exe program to produce a graphical image
   of the {@link Tree} data structure.
<p>
   See
   <br>
   <a href="http://www.graphviz.org/doc/info/lang.html" target="_top">
            http://www.graphviz.org/doc/info/lang.html</a>
   <br>
   and
   <br>
   <a href="http://www.graphviz.org/documentation/" target="_top">
            http://www.graphviz.org/documentation/</a>
*/

public class Tree2dot
{
   private static int nodeCount;

   /**
      Use the {@link String} baseName as the base for the .dot
      file that this program produces from the {@link Tree}
      data structure.

      @param <T>       The element type for the {@link Tree}
      @param tree      {@link Tree} to convert into a dot file
      @param baseName  the base name for the created dot file
   */
   public static <T> void tree2dot(final Tree<T> tree, final String baseName)
   {
      // Create a dot file from a Tree.
      try
      {
         // Create an empty dot file.
         final PrintWriter out = new PrintWriter(
                                    new File(baseName + ".dot"));
         // Put dot commands into the dot file.
         out.println( tree2dot(tree) + "\n" );
         out.close();
      }
      catch (Exception e)
      {
         System.out.println( e );
      }
   }


   /**
      Return a {@link String} representation of the
      dot commands that describe the given {@link Tree}.

      @param <T>   The element type for the {@link Tree}
      @param tree  {@link Tree} to convert into a {@link String} of dot commands
      @return a {@link String} of dot commands representing {@code tree}
   */
   public static <T> String tree2dot(final Tree<T> tree)
   {
      String result = "graph {\n";
      nodeCount = 0;
      // Traverse the Tree and convert it to appropriate dot commands.
      result += tree2dot(tree, nodeCount);
      result += "}\n";
      return result;
   }


   /**
      This method is essentially a pre-order traversal
      of the tree.

      @param <T>    The element type for the {@link Tree}
      @param tree   {@link Tree} to convert into a {@link String} of dot commands
      @param nodeNumber  a number that represents the current node of the tree
      @return a {@link String} of dot commands representing {@code tree}
   */
   public static <T> String tree2dot(final Tree<T> tree, final int nodeNumber)
   {
      String result = "node" + nodeNumber + "[label=\"" + tree.root() + "\"];\n";

      // recursively traverse all the sub trees
      for ( final Tree<T> subTree : tree.subTrees() )
      {
         // create an edge to the sub tree
         result += "node" + nodeNumber + " -- " + "node" + (++nodeCount) + ";\n";
         // convert the sub tree into a dot description
         result += tree2dot(subTree, nodeCount);
      }

      return result;
   }
}//Tree2dot
