/*

*/

package tree;

import java.io.PrintWriter;
import java.io.File;

/**
   This program converts a {@link Tree} of {@link String}s
   into an XML language description of the {@link Tree}.
   The XML description is written to a {@link File} with
   the .xml extension. The .xml file can then be opened
   with an XML viewer (for example, any web browser) to
   produce a graphical representation of the {@link Tree}
   data structure.
   <p>
   This conversion will fail if any string in the tree
   contains one of XML's five reserved characters,
   {@code <, >, &, ", or '}.
<p>
   See
   <br>
   <a href="https://en.wikipedia.org/wiki/XML" target="_top">
            https://en.wikipedia.org/wiki/XML</a>
*/

public class Tree2xml
{
   /**
      Use the {@link String} {@code baseName} as the base for
      the .xml file that this program produces from the
      {@link Tree} of {@link String}s.

      @param tree      {@link Tree} of {@link String}s to convert into an xml file
      @param baseName  the base name for the created xml file
   */
   public static void tree2xml(final Tree<String> tree, final String baseName)
   {
      // Create an .xml file from a tree.
      try
      {
         // Create an empty .xml file.
         final PrintWriter out = new PrintWriter(
                                    new File(baseName + ".xml"));
         // Put xml code into the .xml file.
         out.println("<?xml version=\"1.0\" ?>");
         out.println("<AbstractSyntaxTree>");
         // Traverse the tree and convert it to appropriate xml tags.
         out.println( tree2xml(tree) );
         out.println("</AbstractSyntaxTree>");
         out.close();
      }
      catch (Exception e)
      {
         System.out.println( e );
      }
   }


   /**
      Return a {@link String} representation of the
      xml tags that describes the given {@link Tree}
      of {@link String}s.
      <p>
      This method is essentially a pre-order traversal
      of the tree.

      @param tree  {@link Tree} of {@link String}s to convert into an xml {@link String}
      @return a {@link String} of xml tags representing the abstract syntax tree
   */
   public static String tree2xml(final Tree<String> tree)
   {
      String result = "<tree>";
      result += "<root>";
      result += tree.root();
      result += "</root>";
      // Recursively traverse all the sub trees.
      for ( final Tree<String> subTree : tree.subTrees() )
      {
         // Convert the sub tree into an xml description.
         result += tree2xml(subTree);
      }
      result += "</tree>";

      return result;
   }
}
